// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

#define dos_warn
double max=-1000;
double min=1000;

static struct dos dosn;
static struct dos dosp;

double dos_srh_get_fermi_n(double n, double p,int band)
{
double srh_sigman=dosn.config.srh_sigman;
double srh_sigmap=dosn.config.srh_sigmap;
double Nc=dosn.config.Nc;
double Nv=dosn.config.Nv;
double srh_vth=dosn.config.srh_vth;
//double srh_Nt=dosn.srh_den[band];
double srh_en=srh_vth*srh_sigman*Nc*exp((q*dosn.srh_E[band])/(300.0*kb));
double srh_ep=srh_vth*srh_sigmap*Nv*exp((q*(-1.0-dosn.srh_E[band]))/(300.0*kb));
//printf("%le\n",dosn.srh_E[band]);
//getchar();
double f=0.0;
f=(n*srh_vth*srh_sigman+srh_ep)/(n*srh_vth*srh_sigman+p*srh_vth*srh_sigmap+srh_en+srh_ep);
double level=0.0;
level=dosn.srh_E[band]-300.0*kb*log((1.0/f)-1.0)/q;
return level;
}

double dos_srh_get_fermi_p(double n, double p,int band)
{
double srh_sigmap=dosp.config.srh_sigmap;
double srh_sigman=dosp.config.srh_sigman;
double Nc=dosp.config.Nc;
double Nv=dosp.config.Nv;
double srh_vth=dosp.config.srh_vth;
//double srh_Nt=dosn.srh_den[band];
double srh_ep=srh_vth*srh_sigmap*Nv*exp((q*dosp.srh_E[band])/(300.0*kb));
double srh_en=srh_vth*srh_sigman*Nc*exp((q*(-1.0-dosp.srh_E[band]))/(300.0*kb));
double f=0.0;
f=(p*srh_vth*srh_sigmap+srh_en)/(p*srh_vth*srh_sigmap+n*srh_vth*srh_sigman+srh_ep+srh_en);
double level=0.0;
level=dosp.srh_E[band]-300.0*kb*log((1.0/f)-1.0)/q;
return level;
}

int search(double *x,int N,double find)
{
if (N==1) return 0;
int pos=N/2;
int step=N/2;
do
{
	step=step/2 + (step % 2 > 0 ? 1 : 0);

	if (x[pos]>find)
	{
		pos-=step;
	}else
	{
		pos+=step;
	}

	if (pos<=0)
	{
		pos=0;
		break;
	}
	if (pos>=(N-1))
	{
		pos=N-1;
		break;
	}
	if (step==0) break;
	if (x[pos]==find) break;
	if ((x[pos]<=find)&&((x[pos+1]>find))) break;

}while(1);

if (pos==(N-1)) pos=N-2;


return pos;
}

void dos_init()
{
dosn.used=FALSE;
dosp.used=FALSE;
}

void dos_free_now(struct dos *mydos)
{
int ii;
int iii;
if (mydos->used==TRUE)
{
	free(mydos->x);
	free(mydos->t);
	free(mydos->srh_E);
	free(mydos->srh_den);

	for (ii=0;ii<mydos->tlen;ii++)
	{
		for (iii=0;iii<mydos->xlen;iii++)
		{
			free(mydos->srh_r1[ii][iii]);
			free(mydos->srh_r2[ii][iii]);
			free(mydos->srh_r3[ii][iii]);
			free(mydos->srh_r4[ii][iii]);
			free(mydos->srh_c[ii][iii]);

		}

		free(mydos->c[ii]);
		free(mydos->w[ii]);
		free(mydos->srh_r1[ii]);
		free(mydos->srh_r2[ii]);
		free(mydos->srh_r3[ii]);
		free(mydos->srh_r4[ii]);
		free(mydos->srh_c[ii]);
	}

	free(mydos->c);
	free(mydos->w);

	free(mydos->srh_r1);
	free(mydos->srh_r2);
	free(mydos->srh_r3);
	free(mydos->srh_r4);
	free(mydos->srh_c);

	mydos->xlen=0;
	mydos->tlen=0;
	mydos->srh_bands=0;
}

}

void dos_free()
{
	dos_free_now(&dosn);
	dos_free_now(&dosp);
}

void load_dos_file(struct dos *mydos,char *file)
{
#ifndef dos_bin
double srh_r1=0.0;
double srh_r2=0.0;
double srh_r3=0.0;
double srh_r4=0.0;
double srh_c=0.0;
double w0;
double n;
#endif
mydos->used=TRUE;
mydos->used=TRUE;

printf("Loading %s\n",file);

	FILE *in;
	/*FILE *rt;
		#ifdef dos_bin
		if (strcmp(file,"./dosn.dat")==0)
		{
		rt = fopen("./rtbinn.inp", "w");
		}else
		{
		rt = fopen("./rtbinp.inp", "w");
		}
		#else
		if (strcmp(file,"./dosn.dat")==0)
		{
		rt = fopen("./rtnormn.inp", "w");
		}else
		{
		rt = fopen("./rtnormp.inp", "w");
		}
		#endif*/
	#ifdef dos_bin
	long len;
	in = fopen(file, "rb");

		if (in==NULL)
		{
			printf("DOS file not found\n");
		}

		fseek(in, 0, SEEK_END);
		len=ftell(in);
		fseek(in, 0, SEEK_SET);
		int buf_len=len/sizeof(double);

		double *buf=(double *)malloc(sizeof(double)*buf_len);


		int buf_pos=0;
		fread((char*)buf, len, 1, in);
		fclose(in);
	#else
		in=fopen(file,"r");
		if (in==NULL)
		{
			printf("DOS n file not found\n");
		}
	#endif


	int t=0;
	int x=0;
	int srh_band=0;
	#ifdef dos_bin
	mydos->xlen=(int)buf[buf_pos++];
	mydos->tlen=(int)buf[buf_pos++];
	mydos->srh_bands=(int)buf[buf_pos++];
	mydos->config.sigma=buf[buf_pos++];
	mydos->config.mu=buf[buf_pos++];
	mydos->config.srh_vth=buf[buf_pos++];
	mydos->config.srh_sigman=buf[buf_pos++];
	mydos->config.srh_sigmap=buf[buf_pos++];
	mydos->config.Nc=buf[buf_pos++];
	mydos->config.Nv=buf[buf_pos++];

	#else
	fscanf(in,"%d %d %d %le %le %le %le %le %le %le\n",&(mydos->xlen),&(mydos->tlen),&(mydos->srh_bands),&(mydos->config.sigma),&(mydos->config.mu),&(mydos->config.srh_vth),&(mydos->config.srh_sigman),&(mydos->config.srh_sigmap),&(mydos->config.Nc),&(mydos->config.Nv));
	#endif
	//fprintf(rt,"%d %d %d %lf %le %le %le %le %le %le\n",(mydos->xlen),(mydos->tlen),(mydos->srh_bands),(mydos->config.sigma),(mydos->config.mu),(mydos->config.srh_vth),(mydos->config.srh_sigman),(mydos->config.srh_sigmap),(mydos->config.Nc),(mydos->config.Nv));
	double xsteps=mydos->xlen;
	double tsteps=mydos->tlen;
	mydos->x=(double *)malloc(sizeof(double)*(int)xsteps);
	mydos->t=(double *)malloc(sizeof(double)*(int)tsteps);
	mydos->srh_E=(double *)malloc(sizeof(double)*(int)(mydos->srh_bands));
	mydos->srh_den=(double *)malloc(sizeof(double)*(int)(mydos->srh_bands));

	mydos->srh_r1=(double ***)malloc(sizeof(double **)*(int)mydos->tlen);
	mydos->srh_r2=(double ***)malloc(sizeof(double **)*(int)mydos->tlen);
	mydos->srh_r3=(double ***)malloc(sizeof(double **)*(int)mydos->tlen);
	mydos->srh_r4=(double ***)malloc(sizeof(double **)*(int)mydos->tlen);
	mydos->srh_c=(double ***)malloc(sizeof(double **)*(int)mydos->tlen);

	mydos->w=(double **)malloc(sizeof(double **)*(int)mydos->tlen);
	mydos->c=(double **)malloc(sizeof(double *)*(int)mydos->tlen);

	for (t=0;t<tsteps;t++)
	{
		mydos->c[t]=(double *)malloc(sizeof(double )*(int)xsteps);
		mydos->w[t]=(double *)malloc(sizeof(double )*(int)xsteps);
		mydos->srh_r1[t]=(double **)malloc(sizeof(double *)*(int)xsteps);
		mydos->srh_r2[t]=(double **)malloc(sizeof(double *)*(int)xsteps);
		mydos->srh_r3[t]=(double **)malloc(sizeof(double *)*(int)xsteps);
		mydos->srh_r4[t]=(double **)malloc(sizeof(double *)*(int)xsteps);
		mydos->srh_c[t]=(double **)malloc(sizeof(double *)*(int)xsteps);

		for (x=0;x<xsteps;x++)
		{
			mydos->srh_r1[t][x]=(double *)malloc(sizeof(double )*(int)(mydos->srh_bands));
			mydos->srh_r2[t][x]=(double *)malloc(sizeof(double )*(int)(mydos->srh_bands));
			mydos->srh_r3[t][x]=(double *)malloc(sizeof(double )*(int)(mydos->srh_bands));
			mydos->srh_r4[t][x]=(double *)malloc(sizeof(double )*(int)(mydos->srh_bands));
			mydos->srh_c[t][x]=(double *)malloc(sizeof(double )*(int)(mydos->srh_bands));
		}

	}

	for (x=0;x<xsteps;x++)
	{
		#ifdef dos_bin
		mydos->x[x]=buf[buf_pos++];
		#else
		fscanf(in,"%le",&(mydos->x[x]));
		#endif
		//fprintf(rt,"%le\n",(mydos->x[x]));
	}

	for (t=0;t<tsteps;t++)
	{
		#ifdef dos_bin
		mydos->t[t]=buf[buf_pos++];
		#else
		fscanf(in,"%le",&(mydos->t[t]));
		#endif
		//fprintf(rt,"%le\n",(mydos->t[t]));
	}

	for (srh_band=0;srh_band<(mydos->srh_bands);srh_band++)
	{
		#ifdef dos_bin
		mydos->srh_E[srh_band]=buf[buf_pos++];
		#else
		fscanf(in,"%le",&(mydos->srh_E[srh_band]));
		#endif
		//fprintf(rt,"%le\n",(mydos->srh_E[srh_band]));
	}

	for (srh_band=0;srh_band<(mydos->srh_bands);srh_band++)
	{
		#ifdef dos_bin
		mydos->srh_den[srh_band]=buf[buf_pos++];
		#else
		fscanf(in,"%le",&(mydos->srh_den[srh_band]));
		#endif
		//fprintf(rt,"%le\n",(mydos->srh_den[srh_band]));
	}

	for (t=0;t<tsteps;t++)
	{
		for (x=0;x<xsteps;x++)
		{

			#ifdef dos_bin
			mydos->c[t][x]=buf[buf_pos++];
			mydos->w[t][x]=buf[buf_pos++];
			#else
			fscanf(in,"%le %le ",&n,&w0);
			mydos->c[t][x]=n;
			mydos->w[t][x]=w0;
			#endif
			//fprintf(rt,"%.20le %.20le ",mydos->c[t][x],mydos->w[t][x]);

			for (srh_band=0;srh_band<mydos->srh_bands;srh_band++)
			{
				#ifdef dos_bin
				mydos->srh_r1[t][x][srh_band]=buf[buf_pos++];
				mydos->srh_r2[t][x][srh_band]=buf[buf_pos++];
				mydos->srh_r3[t][x][srh_band]=buf[buf_pos++];
				mydos->srh_r4[t][x][srh_band]=buf[buf_pos++];
				mydos->srh_c[t][x][srh_band]=buf[buf_pos++];
				//printf("%le\n",mydos->srh_c[t][x][srh_band]);
				#else
				fscanf(in,"%le %le %le %le %le ",&srh_r1,&srh_r2,&srh_r3,&srh_r4,&srh_c);
				mydos->srh_r1[t][x][srh_band]=srh_r1;
				mydos->srh_r2[t][x][srh_band]=srh_r2;
				mydos->srh_r3[t][x][srh_band]=srh_r3;
				mydos->srh_r4[t][x][srh_band]=srh_r4;
				mydos->srh_c[t][x][srh_band]=srh_c;
				#endif
				//fprintf(rt,"%.20le %.20le %.20le %.20le %.20le",mydos->srh_r1[t][x][srh_band],mydos->srh_r2[t][x][srh_band],mydos->srh_r3[t][x][srh_band],mydos->srh_r4[t][x][srh_band],mydos->srh_c[t][x][srh_band]);

			}
		//fprintf(rt,"\n");
		}

	}
	#ifdef dos_bin
	free(buf);
	#else
	fclose(in);
	#endif

//fclose(rt);

}

double dos_get_band_energy_n(int band)
{
return dosn.srh_E[band];
}

double dos_get_band_energy_p(int band)
{
return dosp.srh_E[band];
}

void load_dos(struct device *dev,char *namen, char *namep)
{
load_dos_file(&dosn,namen);
load_dos_file(&dosp,namep);
dev->srh_bands=dosn.srh_bands;
}

double get_n_w(double top,double T,double m,int type)
{
double ret=0.0;
if (type==0)
{
	ret=(3.0/2.0)*kb*T;
}else
{
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;


int t;
int x;

#ifdef dos_warn
if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosn.t,dosn.tlen,T);
x=hashget(dosn.x,dosn.xlen,top);


if (x<0) x=0;
if (t<0) t=0;

x0=dosn.x[x];
x1=dosn.x[x+1];
xr=(top-x0)/(x1-x0);

if (dosn.tlen>1)
{
t0=dosn.t[t];
t1=dosn.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}

c00=dosn.w[t][x];
c01=dosn.w[t][x+1];
c0=c00+xr*(c01-c00);

if (dosn.tlen>1)
{
c10=dosn.w[t+1][x];
c11=dosn.w[t+1][x+1];
c1=c10+xr*(c11-c10);
}

c=c0+tr*(c1-c0);
ret=c;
}
//printf("%le %le\n",q*(3.0/2.0)*ret,(3.0/2.0)*kb*T);
//getchar();
//ret=q*(3.0/2.0)*ret;	//using lookup
ret=(3.0/2.0)*kb*T;//*ret;
return ret;
}

double get_p_w(double top,double T,double m,int type)
{

if (top>max) max=top;
if (top<min) min=top;
//printf("%e %e\n",min,max);

double ret=0.0;
if (type==0)
{
	ret=(3.0/2.0)*kb*T;
}else
{
	double c0=0.0;
	double c1=0.0;
	double x0=0.0;
	double x1=0.0;
	double t0=0.0;
	double t1=0.0;
	double c=0.0;
	double xr=0.0;
	double tr=0.0;
	double c00=0.0;
	double c01=0.0;
	double c10=0.0;
	double c11=0.0;


	int t;
	int x;

	#ifdef dos_warn
	if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
	{
		printf("Holes Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
		//exit(0);
	}
	#endif

/*	for (t=0;t<dosp.tlen-1;t++)
	{
	if (dosp.t[t]>T) break;
	}

	for (x=0;x<dosp.xlen-1;x++)
	{
	if (dosp.x[x]>top) break;
	}
	x--;
	t--;


	if (x<0) x=0;
	if (t<0) t=0;*/

	t=hashget(dosp.t,dosp.tlen,T);
	x=hashget(dosp.x,dosp.xlen,top);

	x0=dosp.x[x];
	x1=dosp.x[x+1];

	xr=(top-x0)/(x1-x0);
	if (xr>1) xr=1;
	t0=dosp.t[t];
	if (dosp.tlen>1)
	{
	t1=dosp.t[t+1];
	tr=(T-t0)/(t1-t0);
	}else
	{
	tr=0.0;
	}

	c00=dosp.w[t][x];
	c01=dosp.w[t][x+1];
	c0=c00+xr*(c01-c00);

	if (dosp.tlen>1)
	{
	c10=dosp.w[t+1][x];
	c11=dosp.w[t+1][x+1];
	c1=c10+xr*(c11-c10);
	}

	c=c0+tr*(c1-c0);
	ret=c;
}

//getchar();
//ret=q*(3.0/2.0)*ret;	//using lookup
ret=(3.0/2.0)*kb*T;//*ret;//kb*T;//get_p_den(top,T, m,type,dos_all)/get_dp_den(top,T, m,type,dos_all);
//printf("%e %e\n",ret,);

return ret;
}

double get_top_from_n(double n,double T,double m,double Nc,int type)
{

double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double cll=0.0;
int xx=0;
double xr=0.0;

int t=0;
int x=0;
//type=0;
if (type==0)
{
	ret=(kb*T/q)*log((fabs(n))/Nc);
}else
{


if (dosn.tlen>1)
{
	for (t=0;t<dosn.tlen-1;t++)
	{
		if (dosn.t[t]>T) break;
	}
	t--;

}else
{
	t=0;
}

xx=0;




//}else
//{
	//t=0;
//}

for (x=0;x<dosn.xlen-1;x++)
{
	if (dosn.c[t][x]>n) break;
}


x--;


//printf("%d %d %d\n",x,xx,dosn.xlen);
//getchar();
if (x<0) x=0;
if (xx<0) xx=0;
if (t<0) t=0;

x0=dosn.c[t][x];
x1=dosn.c[t][x+1];
xr=(n-x0)/(x1-x0);
if (xr>1) xr=1;
c0=dosn.x[x];
c1=dosn.x[x+1];

cll=c0+xr*(c1-c0);
//printf("cll = %e\n",cll);
//if (dosn.tlen>1)
//{

//x0=dosn.c[t+1][xx];
//x1=dosn.c[t+1][xx+1];
//printf("values= %d %d %d\n",xx,dosn.xlen);

/*if (x0==x1)
{
	xr=0;
}else
{
	xr=(n-x0)/(x1-x0);
}
c0=dosn.x[xx];
c1=dosn.x[xx+1];
crr=c0+xr*(c1-c0);
*/
//printf("crr = %e %e %e %e %e %e %e\n",cll,crr,xr,(crr-cll),(n-x0),dosn.c[t+1][xx+1],dosn.c[t+1][xx]);


//	c=cll+(crr-cll)*(T-dosn.t[t])/(dosn.t[t+1]-dosn.t[t]);
//}else
//{
	ret=cll;
//}

//ret=c;

}
//printf("ret = %e\n",ret);
return ret;
}


double get_top_from_p(double p,double T,double m,double Nc,int type)
{
double ret=0.0;

if (type==0)
{
	ret=(kb*T/q)*log((fabs(p))/Nc);
}else
{
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double cll=0.0;
int xx=0;
double xr=0.0;

double c;
int t;
int x;

//if  (dosp.tlen!=1)
//{
	for (t=0;t<dosp.tlen-1;t++)
	{
	if (dosp.t[t]>T) break;
	}
	t--;

	if (t<0) t=0;

	//for (xx=0;xx<dosp.xlen-1;xx++)
	//{
	//if (dosp.c[t+1][xx]>p) break;

	//}
	//xx--;

//}else
//{
	//t=0;
	xx=0;
//}

for (x=0;x<dosp.xlen-1;x++)
{
if (dosp.c[t][x]>p) break;
}
x--;
	//printf("Found between %e %e %e\n",p,dosp.c[t][x],dosp.c[t][x+1]);



if (x<0) x=0;
if (xx<0) xx=0;

x0=dosp.c[t][x];
x1=dosp.c[t][x+1];
xr=(p-x0)/(x1-x0);
c0=dosp.x[x];
c1=dosp.x[x+1];
if (xr>1) xr=1;
cll=c0+xr*(c1-c0);

	//if (dosp.tlen>1)
	//{
	//	x0=dosp.c[t+1][xx];
	//	x1=dosp.c[t+1][xx+1];
	//	xr=(p-x0)/(x1-x0);
	//	c0=dosp.x[xx];
	//	c1=dosp.x[xx+1];
	//	crr=c0+xr*(c1-c0);

	//	c=cll+(crr-cll)*(T-dosp.t[t])/(dosp.t[t+1]-dosp.t[t]);
	//}else
	//{
		c=cll;
	//}
ret=c;
}
return ret;
}

double get_n_den(double top,double T,double m,int type)
{
double ret=0.0;

//type=0;
if (type==0)
{
	double N=2.0*pow(((2.0*pi*kb*T*m*m0)/(hp*hp)),1.5);
	ret=N*exp((top*q)/(kb*T));
}else
{
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;
int t=0;
int x=0;

#ifdef dos_warn
if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
	//getchar();
	//exit(0);
}
#endif

t=hashget(dosn.t,dosn.tlen,T);
//t=search(dosn.t,dosn.tlen,T);
x=hashget(dosn.x,dosn.xlen,top);

if (x<0) x=0;
if (t<0) t=0;

x0=dosn.x[x];
x1=dosn.x[x+1];
xr=(top-x0)/(x1-x0);

if (dosn.tlen>1)
{
t0=dosn.t[t];
t1=dosn.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}

c00=dosn.c[t][x];
c01=dosn.c[t][x+1];
c0=c00+xr*(c01-c00);

if (dosn.tlen>1)
{
c10=dosn.c[t+1][x];
c11=dosn.c[t+1][x+1];
c1=c10+xr*(c11-c10);
}

c=c0+tr*(c1-c0);

ret=c;
}

	//double N=2.0*pow(((2.0*pi*kb*T*m*m0)/(hp*hp)),1.5);
	//double test=N*exp((top*q)/(kb*T));
	//printf("test = %e %e\n",test,ret);
	//getchar();
return ret;
}



double get_n_srh(double top,double T,int trap,int r)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;
int t=0;
int x=0;

#ifdef dos_warn
if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosn.t,dosn.tlen,T);
x=hashget(dosn.x,dosn.xlen,top);

if (x<0) x=0;
if (t<0) t=0;

x0=dosn.x[x];
x1=dosn.x[x+1];
xr=(top-x0)/(x1-x0);
if (dosn.tlen>1)
{
t0=dosn.t[t];
t1=dosn.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}

if (r==1)
{
	c00=dosn.srh_r1[t][x][trap];
	c01=dosn.srh_r1[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r1[t+1][x][trap];
		c11=dosn.srh_r1[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}else
if (r==2)
{
	c00=dosn.srh_r2[t][x][trap];
	c01=dosn.srh_r2[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r2[t+1][x][trap];
		c11=dosn.srh_r2[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}else
if (r==3)
{
	c00=dosn.srh_r3[t][x][trap];
	c01=dosn.srh_r3[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r3[t+1][x][trap];
		c11=dosn.srh_r3[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}else
if (r==4)
{
	c00=dosn.srh_r4[t][x][trap];
	c01=dosn.srh_r4[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r4[t+1][x][trap];
		c11=dosn.srh_r4[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}

c=c0+tr*(c1-c0);
ret=c;

return ret;
}

double get_p_srh(double top,double T,int trap,int r)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;
int t=0;
int x=0;

#ifdef dos_warn
if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosp.t,dosp.tlen,T);
x=hashget(dosp.x,dosp.xlen,top);

if (x<0) x=0;
if (t<0) t=0;

x0=dosp.x[x];
x1=dosp.x[x+1];
xr=(top-x0)/(x1-x0);
if (dosp.tlen>1)
{
t0=dosp.t[t];
t1=dosp.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}

if (r==1)
{
	c00=dosp.srh_r1[t][x][trap];
	c01=dosp.srh_r1[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r1[t+1][x][trap];
		c11=dosp.srh_r1[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}else
if (r==2)
{
	c00=dosp.srh_r2[t][x][trap];
	c01=dosp.srh_r2[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r2[t+1][x][trap];
		c11=dosp.srh_r2[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}else
if (r==3)
{
	c00=dosp.srh_r3[t][x][trap];
	c01=dosp.srh_r3[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r3[t+1][x][trap];
		c11=dosp.srh_r3[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}else
if (r==4)
{
	c00=dosp.srh_r4[t][x][trap];
	c01=dosp.srh_r4[t][x+1][trap];
	c0=c00+xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r4[t+1][x][trap];
		c11=dosp.srh_r4[t+1][x+1][trap];
		c1=c10+xr*(c11-c10);
	}
}

c=c0+tr*(c1-c0);
ret=c;

return ret;
}



double get_dn_srh(double top,double T,int trap,int r)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;


int t;
int x;

#ifdef dos_warn
if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosn.t,dosn.tlen,T);
x=hashget(dosn.x,dosn.xlen,top);

x0=dosn.x[x];
x1=dosn.x[x+1];
xr=1.0/(x1-x0);

if (dosn.tlen>1)
{
t0=dosn.t[t];
t1=dosn.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}

if (r==1)
{
	c00=dosn.srh_r1[t][x][trap];
	c01=dosn.srh_r1[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r1[t+1][x][trap];
		c11=dosn.srh_r1[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}else
if (r==2)
{
	c00=dosn.srh_r2[t][x][trap];
	c01=dosn.srh_r2[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r2[t+1][x][trap];
		c11=dosn.srh_r2[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}else
if (r==3)
{
	c00=dosn.srh_r3[t][x][trap];
	c01=dosn.srh_r3[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r3[t+1][x][trap];
		c11=dosn.srh_r3[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}else
if (r==4)
{
	c00=dosn.srh_r4[t][x][trap];
	c01=dosn.srh_r4[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosn.tlen>1)
	{
		c10=dosn.srh_r4[t+1][x][trap];
		c11=dosn.srh_r4[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}

c=c0+tr*(c1-c0);

ret=c;

return ret;
}


double get_dp_srh(double top,double T,int trap,int r)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;


int t;
int x;

#ifdef dos_warn
if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosp.t,dosp.tlen,T);
x=hashget(dosp.x,dosp.xlen,top);

x0=dosp.x[x];
x1=dosp.x[x+1];
xr=1.0/(x1-x0);

if (dosp.tlen>1)
{
t0=dosp.t[t];
t1=dosp.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}

if (r==1)
{
	c00=dosp.srh_r1[t][x][trap];
	c01=dosp.srh_r1[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r1[t+1][x][trap];
		c11=dosp.srh_r1[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}else
if (r==2)
{
	c00=dosp.srh_r2[t][x][trap];
	c01=dosp.srh_r2[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r2[t+1][x][trap];
		c11=dosp.srh_r2[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}else
if (r==3)
{
	c00=dosp.srh_r3[t][x][trap];
	c01=dosp.srh_r3[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r3[t+1][x][trap];
		c11=dosp.srh_r3[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}else
if (r==4)
{
	c00=dosp.srh_r4[t][x][trap];
	c01=dosp.srh_r4[t][x+1][trap];
	c0=xr*(c01-c00);

	if (dosp.tlen>1)
	{
		c10=dosp.srh_r4[t+1][x][trap];
		c11=dosp.srh_r4[t+1][x+1][trap];
		c1=xr*(c11-c10);
	}
}

c=c0+tr*(c1-c0);

ret=c;

return ret;
}



/////////////////////////////////////////////////////trap

double get_n_pop_srh(double top,double T,int trap)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;
int t=0;
int x=0;

#ifdef dos_warn
if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosn.t,dosn.tlen,T);
x=hashget(dosn.x,dosn.xlen,top);

if (x<0) x=0;
if (t<0) t=0;

x0=dosn.x[x];
x1=dosn.x[x+1];
xr=(top-x0)/(x1-x0);
if (dosn.tlen>1)
{
t0=dosn.t[t];
t1=dosn.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}


c00=dosn.srh_c[t][x][trap];
c01=dosn.srh_c[t][x+1][trap];
c0=c00+xr*(c01-c00);

if (dosn.tlen>1)
{
	c10=dosn.srh_c[t+1][x][trap];
	c11=dosn.srh_c[t+1][x+1][trap];
	c1=c10+xr*(c11-c10);
}

c=c0+tr*(c1-c0);
ret=c;

return ret;
}

double get_p_pop_srh(double top,double T,int trap)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;
int t=0;
int x=0;

#ifdef dos_warn
if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosp.t,dosp.tlen,T);
x=hashget(dosp.x,dosp.xlen,top);

if (x<0) x=0;
if (t<0) t=0;

x0=dosp.x[x];
x1=dosp.x[x+1];
xr=(top-x0)/(x1-x0);
if (dosp.tlen>1)
{
t0=dosp.t[t];
t1=dosp.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}


c00=dosp.srh_c[t][x][trap];
c01=dosp.srh_c[t][x+1][trap];
c0=c00+xr*(c01-c00);

if (dosp.tlen>1)
{
	c10=dosp.srh_c[t+1][x][trap];
	c11=dosp.srh_c[t+1][x+1][trap];
	c1=c10+xr*(c11-c10);
}


c=c0+tr*(c1-c0);
ret=c;

return ret;
}



double get_dn_pop_srh(double top,double T,int trap)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;


int t;
int x;

#ifdef dos_warn
if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosn.t,dosn.tlen,T);
x=hashget(dosn.x,dosn.xlen,top);

x0=dosn.x[x];
x1=dosn.x[x+1];
xr=1.0/(x1-x0);

if (dosn.tlen>1)
{
t0=dosn.t[t];
t1=dosn.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}


c00=dosn.srh_c[t][x][trap];
c01=dosn.srh_c[t][x+1][trap];
c0=xr*(c01-c00);

if (dosn.tlen>1)
{
	c10=dosn.srh_c[t+1][x][trap];
	c11=dosn.srh_c[t+1][x+1][trap];
	c1=xr*(c11-c10);
}


c=c0+tr*(c1-c0);

ret=c;

return ret;
}


double get_dp_pop_srh(double top,double T,int trap)
{
double ret=0.0;
double c0=0.0;
double c1=0.0;
double x0=0.0;
double x1=0.0;
double t0=0.0;
double t1=0.0;
double c=0.0;
double xr=0.0;
double tr=0.0;
double c00=0.0;
double c01=0.0;
double c10=0.0;
double c11=0.0;


int t;
int x;

#ifdef dos_warn
if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
{
	printf("Electrons Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
	//exit(0);
}
#endif

t=hashget(dosp.t,dosp.tlen,T);
x=hashget(dosp.x,dosp.xlen,top);

x0=dosp.x[x];
x1=dosp.x[x+1];
xr=1.0/(x1-x0);

if (dosp.tlen>1)
{
t0=dosp.t[t];
t1=dosp.t[t+1];
tr=(T-t0)/(t1-t0);
}else
{
tr=0.0;
}


c00=dosp.srh_c[t][x][trap];
c01=dosp.srh_c[t][x+1][trap];
c0=xr*(c01-c00);

if (dosp.tlen>1)
{
	c10=dosp.srh_c[t+1][x][trap];
	c11=dosp.srh_c[t+1][x+1][trap];
	c1=xr*(c11-c10);
}


c=c0+tr*(c1-c0);

ret=c;

return ret;
}

/////////////////////////////////////////////////////trap


double get_n_mu()
{
return dosn.config.mu;
}


double get_dn_den(double top,double T,double m,int type)
{

double ret=0.0;
if (type==0)
{
	double N=2.0*pow(((2.0*pi*kb*T*m*m0)/(hp*hp)),1.5);
	ret=(q/(kb*T))*N*exp((top*q)/(kb*T));
}else
{
	double c0=0.0;
	double c1=0.0;
	double x0=0.0;
	double x1=0.0;
	double t0=0.0;
	double t1=0.0;
	double c=0.0;
	double xr=0.0;
	double tr=0.0;
	double c00=0.0;
	double c01=0.0;
	double c10=0.0;
	double c11=0.0;


	int t;
	int x;

	#ifdef dos_warn
	if ((dosn.x[0]>top)||(dosn.x[dosn.xlen-1]<top))
	{
		printf("Electrons Asking for %e but range %e %e\n",top,dosn.x[0],dosn.x[dosn.xlen-1]);
		//exit(0);
	}
	#endif

	t=hashget(dosn.t,dosn.tlen,T);
	x=hashget(dosn.x,dosn.xlen,top);

	x0=dosn.x[x];
	x1=dosn.x[x+1];
	xr=1.0/(x1-x0);

	if (dosn.tlen>1)
	{
	t0=dosn.t[t];
	t1=dosn.t[t+1];
	tr=(T-t0)/(t1-t0);
	}else
	{
	tr=0.0;
	}
//printf("%d %d %d %d\n",t,x,dosn.tlen,dosn.xlen);
	c00=dosn.c[t][x];
	c01=dosn.c[t][x+1];
	c0=xr*(c01-c00);

	if (dosn.tlen>1)
	{
	c10=dosn.c[t+1][x];
	c11=dosn.c[t+1][x+1];
	c1=xr*(c11-c10);
	}

	c=c0+tr*(c1-c0);

	ret=c;
}

return ret;
}

void draw_gaus(struct device *in)
{
/*
FILE *out=fopen("./gaus.dat","w");
double dE=1e-3;
double E=in->Ev[0]-1;
double Ev=in->Ev[0];
double Ec=in->Ec[0];
double Estop=in->Ec[0]+1;
double gauEv;
double gauEc;
double sigmae=dosn[1].config.sigma;
double sigmah=dosp[1].config.sigma;
//printf("sigmas %e %e\n",fabs(dos_get_Ev_edge(in->dostype[0])),fabs(dos_get_Ec_edge(in->dostype[0])));
//getchar();
//getchar();
do
{
gauEv=(1.0/(sqrt(2.0*pi)*sigmae))*exp(-pow(((E-Ev)/(sqrt(2.0)*sigmae)),2.0));
gauEc=(1.0/(sqrt(2.0*pi)*sigmah))*exp(-pow(((E-Ec)/(sqrt(2.0)*sigmah)),2.0));
fprintf(out,"%e %e\n",E,gauEc+gauEv);
E+=dE;
}
while(E<Estop);
fclose(out);*/
}




double get_p_den(double top,double T,double m,int type)
{

if (top>max) max=top;
if (top<min) min=top;
//printf("%e %e\n",min,max);

double ret=0.0;
if (type==0)
{
	double N=2.0*pow(((2.0*pi*kb*T*m*m0)/(hp*hp)),1.5);
	ret=N*exp((top*q)/(kb*T));
}else
{
	double c0=0.0;
	double c1=0.0;
	double x0=0.0;
	double x1=0.0;
	double t0=0.0;
	double t1=0.0;
	double c=0.0;
	double xr=0.0;
	double tr=0.0;
	double c00=0.0;
	double c01=0.0;
	double c10=0.0;
	double c11=0.0;


	int t;
	int x;

	#ifdef dos_warn
	if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
	{
		printf("Holes Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
		//exit(0);
	}
	#endif

	t=hashget(dosp.t,dosp.tlen,T);
	x=hashget(dosp.x,dosp.xlen,top);
	//if ((x<0)||(x>=dosp.xlen-10)) printf("%d %d\n",x,dosp.xlen);
	x0=dosp.x[x];
	x1=dosp.x[x+1];
	xr=(top-x0)/(x1-x0);


	if (dosp.tlen>1)
	{
	t0=dosp.t[t];
	t1=dosp.t[t+1];
	tr=(T-t0)/(t1-t0);
	}else
	{
	tr=0.0;
	}

	c00=dosp.c[t][x];
	c01=dosp.c[t][x+1];
	c0=c00+xr*(c01-c00);

	if (dosp.tlen>1)
	{
	c10=dosp.c[t+1][x];
	c11=dosp.c[t+1][x+1];
	c1=c10+xr*(c11-c10);
	}

	c=c0+tr*(c1-c0);
	ret=c;
}

return ret;
}

double get_p_mu(double top,double T,double m,int type,int timedomain)
{
return dosp.config.mu;

}

double get_dp_den(double top,double T,double m,int type)
{

double ret=0.0;
if (type==0)
{
	double N=2.0*pow(((2.0*pi*kb*T*m*m0)/(hp*hp)),1.5);
	ret=(q/(kb*T))*N*exp((top*q)/(kb*T));
}else
{
	double c0=0.0;
	double c1=0.0;
	double x0=0.0;
	double x1=0.0;
	double t0=0.0;
	double t1=0.0;
	double c=0.0;
	double xr=0.0;
	double tr=0.0;
	double c00=0.0;
	double c01=0.0;
	double c10=0.0;
	double c11=0.0;


	int t;
	int x;

	#ifdef dos_warn
	if ((dosp.x[0]>top)||(dosp.x[dosp.xlen-1]<top))
	{
		printf("Holes Asking for %e but range %e %e\n",top,dosp.x[0],dosp.x[dosp.xlen-1]);
		//exit(0);
	}
	#endif

	t=hashget(dosp.t,dosp.tlen,T);
	x=hashget(dosp.x,dosp.xlen,top);

	x0=dosp.x[x];
	x1=dosp.x[x+1];
	xr=1.0/(x1-x0);

	if (dosp.tlen>1)
	{
	t0=dosp.t[t];
	t1=dosp.t[t+1];
	tr=(T-t0)/(t1-t0);
	}else
	{
	tr=0.0;
	}

	c00=dosp.c[t][x];
	c01=dosp.c[t][x+1];
	c0=xr*(c01-c00);

	if (dosp.tlen>1)
	{
	c10=dosp.c[t+1][x];
	c11=dosp.c[t+1][x+1];
	c1=xr*(c11-c10);
	}

	c=c0+tr*(c1-c0);
	ret=c;

}


return ret;
}
