// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

static int frame_config_change;
void frame_free(struct map *in)
{

#ifdef dumpvid
int i;

for (i=0;i<in->xpoints;i++)
{
	free(in->data[i]);
}
free(in->data);

free(in->x);
free(in->y);
#endif
}

void frame_init(struct map *in,int xpoints,int ypoints,double xstart,double xstop,double ystart,double ystop)
{
#ifdef dumpvid
char temp[200];
FILE *file=fopen("./dump.inp","r");
fscanf(file,"%s",temp);
fscanf(file,"%d",&frame_config_change);
fclose(file);

int x;
int y;
double pos=0.0;
in->xpoints=xpoints;
in->ypoints=ypoints;
in->xstart=xstart;
in->xstop=xstop;
in->ystart=ystart;
in->ystop=ystop;
in->count=0;

in->data=(double **)malloc (in->xpoints*sizeof(double*));
for (x=0;x<in->xpoints;x++)
{
in->data[x]=(double *)malloc (in->ypoints*sizeof(double));
}

for (x=0;x<in->xpoints;x++)
{
	for (y=0;y<in->ypoints;y++)
	{
		in->data[x][y]=0.0;
	}
}

pos=in->xstart;
in->xdelta=(in->xstop-in->xstart)/((double)in->xpoints);
in->x=(double *)malloc (in->xpoints*sizeof(double));
for (x=0;x<in->xpoints;x++)
{
	in->x[x]=pos;
	pos+=in->xdelta;
}

pos=in->ystart;
in->ydelta=(in->ystop-in->ystart)/((double)in->ypoints);
in->y=(double *)malloc (in->ypoints*sizeof(double));
for (y=0;y<in->ypoints;y++)
{
	in->y[y]=pos;
	pos+=in->ydelta;
}
#endif
}

void frame_reset(struct map *in)
{
#ifdef dumpvid
int x;
int y;
for (x=0;x<in->xpoints;x++)
{
	for (y=0;y<in->ypoints;y++)
	{
		in->data[x][y]=0.0;
	}
}
#endif
}

void frame_build(struct map *mapin,struct device *in)
{
#ifdef dumpvid
int i;
int band;
for (i=0;i<in->ymeshpoints;i++)
{
if (frame_config_change==TRUE)
{
	frame_add_data(mapin,in->ymesh[i],in->Ec[i],in->n[i]-in->nfinit[i]);
	frame_add_data(mapin,in->ymesh[i],in->Ev[i],in->p[i]-in->pfinit[i]);

	for (band=0;band<in->srh_bands;band++)
	{
		frame_add_data(mapin,in->ymesh[i],in->Ec[i]+dos_get_band_energy_n(band),in->nt[i][band]-in->ntbinit[i][band]);
		//printf("%le\n",dos_get_band_energy_p(band));
		//getchar();
		frame_add_data(mapin,in->ymesh[i],in->Ev[i]-dos_get_band_energy_p(band),in->pt[i][band]-in->ptbinit[i][band]);
	}
}else
{
	frame_add_data(mapin,in->ymesh[i],in->Ec[i],in->n[i]);
	frame_add_data(mapin,in->ymesh[i],in->Ev[i],in->p[i]);

	for (band=0;band<in->srh_bands;band++)
	{
		frame_add_data(mapin,in->ymesh[i],in->Ec[i]+dos_get_band_energy_n(band),in->nt[i][band]);
		//printf("%le\n",dos_get_band_energy_p(band));
		//getchar();
		frame_add_data(mapin,in->ymesh[i],in->Ev[i]-dos_get_band_energy_p(band),in->pt[i][band]);
	}
}
}
#endif
}

void frame_dump(struct map *in,struct device *dev)
{


#ifdef dumpvid
int x;
int y;
double ret;
char temp[200];
FILE *out;
out=fopen("./frame.plot","w");
fprintf(out,"set title 'V=%.3lf (V) time=%le (s)'\n",dev->Vapplied,dev->time);
fprintf(out,"set term post enh \"Helvetica\" 20 color\n");
fprintf(out,"set pm3d map\n");
fprintf(out,"set xlabel 'Position (nm)'\n");
fprintf(out,"set ylabel 'Energy (eV)'\n");
fprintf(out,"set logscale cb\n");
//fprintf(out,"set cbrange[1:1e25]\n");
fprintf(out,"set xrange[%lf:%lf]\n",0.0,(1e9*dev->ylen));
fprintf(out,"set format cb \"10^\{%%L}\"\n");
fprintf(out,"splot 'frame.dat' using ($1*1e9):($2):($3) title ''\n");
/*
fprintf(out,"set title 'V=%.3lf (V) time=%le (s)'\n",dev->Vapplied,dev->time);
fprintf(out,"set term post enh \"Helvetica\" 20 color\n");
fprintf(out,"set view 77 , 94 , 1 , 1\n");
fprintf(out,"set xlabel 'Position (nm)'\n");
fprintf(out,"set ylabel 'Energy (eV)'\n");
fprintf(out,"set logscale z\n");
fprintf(out,"set zrange [1e10:1e25]\n");
fprintf(out,"set yrange [-6:-4]\n");
//fprintf(out,"set cbrange[1:1e25]\n");
fprintf(out,"set xrange[%lf:%lf]\n",0.0,(1e9*dev->ylen));
fprintf(out,"splot '2dframen.dat' using ($1*1e9):($2):($3) with l title '','2dframep.dat' using ($1*1e9):($2):($3) with l title ''\n");
*/

fclose(out);

out=fopen("./frame.dat","w");
for (x=0;x<in->xpoints;x++)
{
	for (y=0;y<in->ypoints;y++)
	{
			ret=in->data[x][y];
			//if (ret<1) ret=1;
			if (ret!=0.0)
			fprintf(out,"%le %le %le\n",in->x[x],in->y[y],ret);
	}
fprintf(out,"\n");


}
fclose(out);

//dump_device_map(dev);
//getchar();
//getchar();
system("gnuplot frame.plot >frame.eps");
sprintf(temp,"convert -rotate 90 -resize 800x600 -background white -layers flatten -density 600x600 frame.eps frame.jpg");
printf("running %s\n",temp);
system(temp);
sprintf(temp,"mv frame.eps %d.eps",in->count);
system(temp);


sprintf(temp,"gs  -sDEVICE=jpeg -dJPEGQ=100 -dNOPAUSE -dBATCH -dSAFER -r300 -sOutputFile=./r%d.jpg -f %d.eps",in->count,in->count);
system(temp);

sprintf(temp,"jpegtran -trim -rotate 90  r%d.jpg >%d.jpg",in->count,in->count);
system(temp);

//sprintf(temp,"rm r%d.jpg",in->count);
//sprintf(temp,"rm %d.eps",in->count);
system(temp);
getchar();
in->count++;
#endif


}

void dump_device_map(struct device *in)
{
int i;
int band;
FILE *out;
out=fopen("./2dframen.dat","w");

for (i=0;i<in->ymeshpoints;i++)
{

		for (band=0;band<in->srh_bands;band++)
		{

			fprintf(out,"%le %le %le\n",in->ymesh[i],in->Ec[i]+dos_get_band_energy_n(band),in->nt[i][band]);
//			fprintf(out,"%le %le %le\n",in->ymesh[i],in->Ev[i]-dos_get_band_energy_p(band),in->pt[i][band]);

		}

		fprintf(out,"%le %le %le\n",in->ymesh[i],in->Ec[i],in->n[i]);

		//fprintf(out,"%le %le %le\n",in->ymesh[i],in->Ev[i],in->p[i]);
		fprintf(out,"\n");

}
fclose(out);

out=fopen("./2dframep.dat","w");

for (i=0;i<in->ymeshpoints;i++)
{

		for (band=0;band<in->srh_bands;band++)
		{
			fprintf(out,"%le %le %le\n",in->ymesh[i],in->Ev[i]-dos_get_band_energy_p(band),in->pt[i][band]);

		}
		fprintf(out,"%le %le %le\n",in->ymesh[i],in->Ev[i],in->p[i]);
		fprintf(out,"\n");

}
fclose(out);
}

void frame_add_data(struct map *in,double x,double y,double data)
{
#ifdef dumpvid
int xpos=(x-in->xstart)/in->xdelta;
int ypos=(y-in->ystart)/in->ydelta;
if (xpos<0) xpos=0;
if (ypos<0) ypos=0;
if (xpos>=in->xpoints) xpos=in->xpoints-1;
if (ypos>=in->ypoints) ypos=in->ypoints-1;

//printf("%d %d\n",xpos,ypos);
if (in->data[xpos][ypos]==0)
{
	in->data[xpos][ypos]=data;
}else
{
	in->data[xpos][ypos]=1e25;
}

#endif
}

