// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>
double min_thermal_error=2e-11;

void update_heat(struct device *in)
{
int i;
double Ecl=0.0;
double Ecr=0.0;
double Evl=0.0;
double Evr=0.0;
double yl=0.0;
double yr=0.0;
double dh=0.0;
double Jn=0.0;
double Jp=0.0;

for (i=0;i<in->ymeshpoints;i++)
{
	if (i==0)
	{
		Ecl=-in->Xi[0]-in->Vapplied;
		Evl=-in->Xi[0]-in->Vapplied-in->Eg[0];
		yl=in->ymesh[0]-(in->ymesh[1]-in->ymesh[0]);
	}else
	{
		Ecl=in->Ec[i-1];
		Evl=in->Ev[i-1];
		yl=in->ymesh[i-1];
	}

	if (i==in->ymeshpoints-1)
	{
		Ecr=-in->Xi[i]-in->Vr;
		Evr=-in->Xi[i]-in->Vr-in->Eg[i];
		yr=in->ymesh[i]+(in->ymesh[i]-in->ymesh[i-1]);

	}else
	{
		Ecr=in->Ec[i];
		Evr=in->Ev[i];
		yr=in->ymesh[i];
	}

	dh=yr-yl;
	Jn=in->Jn[i];
	Jp=in->Jp[i];

	in->He[i]=(Ecr-Ecl)*Jn/dh+in->Hasorb[i]/2.0;
	in->Hh[i]=(Evr-Evl)*Jp/dh+in->Hasorb[i]/2.0;
	//printf("%e\n",in->Hasorb[i]);
	//getchar();
	in->ke[i]=(5/2+1.5)*kb*(kb/q)*in->Te[i]*in->mun[i]*in->n[i];
	in->kh[i]=(5/2+1.5)*kb*(kb/q)*in->Th[i]*in->mup[i]*in->p[i];

	//printf("%e\n",in->He[i]);
	//getchar();
}

}

void dump_thermal(struct device *in)
{
int i;
FILE *out;
out=fopen("./thermal.dat","w");
for (i=0;i<in->ymeshpoints;i++)
{
	fprintf(out,"%e %e %e %e %e %e %e %e %e\n",in->ymesh[i],in->Tl[i],in->Te[i],in->Th[i],in->Hl[i],in->He[i],in->Hh[i],in->Hasorb[i],in->kl[i]);
}
fclose(out);
}

double get_thermal_error(struct device *in,double *b)
{
double tot=0.0;
int i;
for (i=0;i<in->ymeshpoints*3;i++)
{
	tot+=mod(b[i]);
}
return tot;
}


int solve_thermal(struct device *in)
{
printf("Solve thermal\n");
int i;

int N=in->ymeshpoints*3-2;		//dTldTl
N+=in->ymeshpoints*3-2;			//dTedTe
N+=in->ymeshpoints*3-2;			//dThdTh
N+=in->ymeshpoints;			//dTedTl
N+=in->ymeshpoints;			//dThdTl
N+=in->ymeshpoints;			//dTldTe
N+=in->ymeshpoints;			//dTldTh


int M=in->ymeshpoints*3;
int *Ti =    malloc(N*sizeof(int));
int *Tj =    malloc(N*sizeof(int));
double *Tx = malloc(N*sizeof(double));
double *b = malloc(M*sizeof(double));


double Tll;
double Tlc;
double Tlr;
double Tel;
double Tec;
double Ter;
double Thl;
double Thc;
double Thr;
double yl;
double yc;
double yr;
double dyl=0.0;
double dyc=0.0;
double dyr=0.0;
int ittr=0;
int pos=0;
double error;
double kll=0.0;
double klc=0.0;
double klr=0.0;
double kl0=0.0;
double kl1=0.0;
double dTll=0.0;
double dTlc=0.0;
double dTlr=0.0;
double dTl=0.0;
double Hl=0.0;

double kel=0.0;
double kec=0.0;
double ker=0.0;
double ke0=0.0;
double ke1=0.0;
double dTel=0.0;
double dTec=0.0;
double dTer=0.0;
double dTe=0.0;
double He=0.0;

double khl=0.0;
double khc=0.0;
double khr=0.0;
double kh0=0.0;
double kh1=0.0;
double dThl=0.0;
double dThc=0.0;
double dThr=0.0;
double dTh=0.0;
double Hh=0.0;
double n=0.0;
double p=0.0;
double tau_e=1e-11;
double tau_h=1e-11;
double dTedTl=0.0;
double dThdTl=0.0;
double dTldTe=0.0;
double dTldTh=0.0;

double Jnl=0.0;
double Jnc=0.0;
double Jnr=0.0;

double Jpl=0.0;
double Jpc=0.0;
double Jpr=0.0;

in->thermal_conv=FALSE;
update_heat(in);

	do
	{

		pos=0;
		for (i=0;i<in->ymeshpoints;i++)
		{


			if (i==0)
			{
				kll=in->kl[0];
				kel=in->ke[0];
				khl=in->kh[0];
				//printf("%d\n",in->Tliso);
				//getchar();
				if (in->Tliso==FALSE)
				{
					Tel=in->Tll;
					Thl=in->Tll;
					Tll=in->Tll;
				}else
				{
					Tll=in->Tl[0];
					Tel=in->Tl[0];
					Thl=in->Tl[0];
				}


				yl=in->ymesh[0]-(in->ymesh[1]-in->ymesh[0]);
				Jnl=in->Jn[0];
				Jpl=in->Jp[0];
			}else
			{
				kll=in->kl[i-1];
				kel=in->ke[i-1];
				khl=in->kh[i-1];
				Tll=in->Tl[i-1];
				Tel=in->Te[i-1];
				Thl=in->Th[i-1];
				yl=in->ymesh[i-1];
				Jnl=in->Jn[i-1];
				Jpl=in->Jp[i-1];
			}

			if (i==(in->ymeshpoints-1))
			{
				klr=in->kl[in->ymeshpoints-1];
				ker=in->ke[in->ymeshpoints-1];
				khr=in->kh[in->ymeshpoints-1];
				Tlr=in->Tlr;
				Ter=in->Tlr;
				Thr=in->Tlr;
				Jnr=in->Jn[in->ymeshpoints-1];
				Jpr=in->Jp[in->ymeshpoints-1];

				yr=in->ymesh[i]+(in->ymesh[i]-in->ymesh[i-1]);
			}else
			{
				klr=in->kl[i+1];
				ker=in->ke[i+1];
				khr=in->kh[i+1];
				Tlr=in->Tl[i+1];
				Ter=in->Te[i+1];
				Thr=in->Th[i+1];
				Jnr=in->Jn[i+1];
				Jpr=in->Jp[i+1];
				yr=in->ymesh[i+1];
			}


			yc=in->ymesh[i];
			dyl=yc-yl;
			dyr=yr-yc;

			dyc=(dyl+dyr)/2.0;

			klc=in->kl[i];
			kec=in->ke[i];
			khc=in->kh[i];

			kl0=(kll+klc)/2.0;
			kl1=(klc+klr)/2.0;

			ke0=(kel+kec)/2.0;
			ke1=(kec+ker)/2.0;

			kh0=(khl+khc)/2.0;
			kh1=(khc+khr)/2.0;


			Tlc=in->Tl[i];
			Tec=in->Te[i];
			Thc=in->Th[i];
			Hl=in->Hl[i];
			He=in->He[i];
			Hh=in->Hh[i];

			n=in->n[i];
			p=in->p[i];

			Jnc=in->Jn[i];
			Jpc=in->Jp[i];




/////////////////Gitter/////////////////////////
			dTll=kl0/dyl/dyc;
			dTlc=-(kl0/dyl/dyc+kl1/dyr/dyc);
			dTlr=kl1/dyr/dyc;

			dTl=dTll*Tll+dTlc*Tlc+dTlr*Tlr+(3.0/2.0)*kb*n*(Tec-Tlc)/tau_e+(3.0/2.0)*kb*p*(Thc-Tlc)/tau_h;

			dTlc+=-(3.0/2.0)*kb*n/tau_e;
			dTlc+=-(3.0/2.0)*kb*p/tau_h;

			dTldTe=(3.0/2.0)*kb*n/tau_e;
			dTldTh=(3.0/2.0)*kb*p/tau_h;

			if ((in->Tliso==TRUE)&&(i==0))
			{
				dTlc+=dTll;
				dTll=0.0;
			}

///////////////////Elektronen////////////////////
			dTel=ke0/dyl/dyc;
			dTec=-(ke0/dyl/dyc+ke1/dyr/dyc);
			dTer=ke1/dyr/dyc;

			dTe=dTel*Tel+dTec*Tec+dTer*Ter-(3.0/2.0)*kb*n*(Tec-Tlc)/tau_e;
			dTec+=-(3.0/2.0)*kb*n/tau_e;
			dTedTl=(3.0/2.0)*kb*n/tau_e;

			double Snl=(5.0/2.0)*(kb/q)*Tel*Jnl;
			double Snr=(5.0/2.0)*(kb/q)*Ter*Jnr;
			double dSndy=-(Snr-Snl)/(dyl+dyr);
			double dSnldTe=(5.0/2.0)*(kb/q)*Jnl/(dyl+dyr);
			double dSnrdTe=-(5.0/2.0)*(kb/q)*Jnr/(dyl+dyr);

			dTel+=dSnldTe;
			dTer+=dSnrdTe;

			if ((in->Tliso==TRUE)&&(i==0))
			{
				dTlc+=dTel;
				dTel=0.0;
			}

////////////////Loch////////////////////////////
			dThl=kh0/dyl/dyc;
			dThc=-(kh0/dyl/dyc+kh1/dyr/dyc);
			dThr=kh1/dyr/dyc;

			dTh=dThl*Thl+dThc*Thc+dThr*Thr-(3.0/2.0)*kb*p*(Thc-Tlc)/tau_h;

			dThc+=-(3.0/2.0)*kb*p/tau_h;
			dThdTl=(3.0/2.0)*kb*p/tau_h;

			double Spl=(5.0/2.0)*(kb/q)*Thl*Jpl;
			double Spr=(5.0/2.0)*(kb/q)*Thr*Jpr;
			double dSpdy=-(Spr-Spl)/(dyl+dyr);
			double dSpldTh=(5.0/2.0)*(kb/q)*Jpl/(dyl+dyr);
			double dSprdTh=-(5.0/2.0)*(kb/q)*Jpr/(dyl+dyr);

			dThl+=dSpldTh;
			dThr+=dSprdTh;

			if ((in->Tliso==TRUE)&&(i==0))
			{
				dTlc+=dThl;
				dThl=0.0;
			}

//////////////Und das Rest//////////////////
			if (i!=0)
			{
				Ti[pos]=i;
				Tj[pos]=i-1;
				Tx[pos]=dTll;
				pos++;

				Ti[pos]=in->ymeshpoints+i;
				Tj[pos]=in->ymeshpoints+i-1;
				Tx[pos]=dTel;
				pos++;

				Ti[pos]=2*in->ymeshpoints+i;
				Tj[pos]=2*in->ymeshpoints+i-1;
				Tx[pos]=dThl;
				pos++;


			}

			//lattice
			Ti[pos]=i;
			Tj[pos]=i;
			Tx[pos]=dTlc;
			pos++;

			Ti[pos]=i;
			Tj[pos]=in->ymeshpoints+i;
			Tx[pos]=dTldTe;
			pos++;

			Ti[pos]=i;
			Tj[pos]=2*in->ymeshpoints+i;
			Tx[pos]=dTldTh;
			pos++;


			//electron
			Ti[pos]=in->ymeshpoints+i;
			Tj[pos]=in->ymeshpoints+i;
			Tx[pos]=dTec;
			pos++;


			Ti[pos]=in->ymeshpoints+i;
			Tj[pos]=i;
			Tx[pos]=dTedTl;
			pos++;

			//Hole
			Ti[pos]=2*in->ymeshpoints+i;
			Tj[pos]=2*in->ymeshpoints+i;
			Tx[pos]=dThc;
			pos++;

			Ti[pos]=2*in->ymeshpoints+i;
			Tj[pos]=i;
			Tx[pos]=dThdTl;
			pos++;



			if (i!=(in->ymeshpoints-1))
			{
				Ti[pos]=i;
				Tj[pos]=i+1;
				Tx[pos]=dTlr;
				pos++;

				Ti[pos]=in->ymeshpoints+i;
				Tj[pos]=in->ymeshpoints+i+1;
				Tx[pos]=dTer;
				pos++;

				Ti[pos]=2*in->ymeshpoints+i;
				Tj[pos]=2*in->ymeshpoints+i+1;
				Tx[pos]=dThr;
				pos++;

			}

			b[i]=-(dTl+Hl);
			b[i+in->ymeshpoints]=-(dTe+dSndy+He);
			b[i+2*in->ymeshpoints]=-(dTh+dSpdy+Hh);

		}

		solver(M,N,Ti,Tj, Tx,b);

		for (i=0;i<in->ymeshpoints;i++)
		{
			double clamp=100.0;
			in->Tl[i]+=b[i]/(1.0+fabs(b[i]/clamp/(300.0*kb/q)));
			in->Te[i]+=b[i+in->ymeshpoints]/(1.0+fabs(b[i+in->ymeshpoints]/clamp/(300.0*kb/q)));
			in->Th[i]+=b[i+2*in->ymeshpoints]/(1.0+fabs(b[i+2*in->ymeshpoints]/clamp/(300.0*kb/q)));

			/*in->Tl[i]+=b[i];
			in->Te[i]+=b[i+in->ymeshpoints];
			in->Th[i]+=b[i+2*in->ymeshpoints];*/
		}


		error=get_thermal_error(in,b);
		printf("Thermal error = %e\n",error);
		if ((ittr<2)&&(error<min_thermal_error)) in->thermal_conv=TRUE;
		ittr++;
		dump_thermal(in);
	}while((ittr<3)&&(error>min_thermal_error));

/*
for (i=0;i<in->ymeshpoints;i++)
{
	if (in->Tl[i]>2000) in->Tl[i]=2000.0;
	if (in->Te[i]>2000) in->Te[i]=2000.0;
	if (in->Th[i]>2000) in->Th[i]=2000.0;

}
*/
dump_thermal(in);

free(Ti);
free(Tj);
free(Tx);
free(b);
return 0;
}

